import os
import os.path as osp

import pandas as pd
import numpy as np
from numcodecs import VLenArray
import zarr

from magenpy.utils.compute_utils import generate_slice_dictionary


def _validate_ld_matrix(ld_mat):
    """
    Takes an `LDMatrix` object and checks its contents for validity.
    Specifically, we check that:
     - The dimensions of the matrix and its associated attributes are matching.
     - The LD boundaries are correct.
     - The masking is working properly.
    :param ld_mat: An instance of `LDMatrix`
    :return: True if `ld_mat` has the correct structure, False otherwise.
    """

    attributes = ['snps', 'a1', 'a2', 'maf', 'bp_position', 'cm_position', 'ld_score']

    for attr in attributes:
        attribute = getattr(ld_mat, attr)
        if attribute is None:
            continue
        if len(attribute) != ld_mat.n_elements:
            raise ValueError("Invalid LD Matrix: Attribute dimensions are not aligned!")

    # Check LD bounds:
    ld_bounds = ld_mat.get_masked_boundaries()

    if ld_bounds.shape != (2, ld_mat.n_elements):
        raise ValueError("Invalid LD Matrix: LD boundaries have the wrong dimensions!")

    ld_block_lengths = ld_bounds[1, :] - ld_bounds[0, :]

    # Iterate over the stored LD data to check its dimensions:
    i = 0

    for i, d in enumerate(ld_mat):
        if len(d) != ld_block_lengths[i]:
            raise ValueError(f"Invalid LD Matrix: Element {i} does not have matching LD boundaries!")

    if i != (ld_mat.n_elements - 1):
        raise ValueError(f"Invalid LD Matrix: Conflicting total number of elements!")

    return True


def move_ld_store(z_arr, target_path, overwrite=True):

    source_path = z_arr.store.dir_path()

    if overwrite or not any(os.scandir(target_path)):
        import shutil
        shutil.rmtree(target_path, ignore_errors=True)
        shutil.move(source_path, target_path)

    return zarr.open(target_path)


def delete_ld_store(z_arr):

    try:
        z_arr.store.rmdir()
    except Exception as e:
        print(e)


def from_plink_ld_bin_to_zarr(bin_file, dir_store, ld_boundaries):
    """
    This method takes an LD binary file from PLINK and converts it to
    a chunked Zarr matrix
    :param bin_file: The path to the LD binary file
    :param dir_store: The directory store where the Zarr array will be stored
    :param ld_boundaries: The boundaries for the desired LD matrix.
    """

    n_rows = ld_boundaries.shape[1]
    avg_ncol = int((ld_boundaries[1, :] - ld_boundaries[0, :]).mean())

    n_chunks = estimate_row_chunk_size(n_rows, avg_ncol)

    if avg_ncol == n_rows:
        z_rag = zarr.open(dir_store,
                          mode='w',
                          shape=(n_rows, n_rows),
                          chunks=n_chunks,
                          dtype=float)
    else:
        z_rag = zarr.open(dir_store,
                          mode='w',
                          shape=n_rows,
                          chunks=n_chunks[:1],
                          dtype=object,
                          object_codec=VLenArray(float))

    chunk_size = z_rag.chunks[0]

    for i in range(int(np.ceil(z_rag.shape[0] / chunk_size))):

        n_chunk_rows = min(chunk_size, n_rows - i*chunk_size)
        ld_chunk = np.fromfile(bin_file,
                               offset=i*chunk_size*n_rows,
                               count=n_chunk_rows*n_rows).reshape(n_chunk_rows, n_rows)

        for j, ld in enumerate(ld_chunk):
            idx = i*chunk_size + j
            start, end = ld_boundaries[:, idx]
            ld_chunk[j] = ld_chunk[j][start: end]

        z_rag[i*chunk_size:(i+1)*chunk_size] = ld_chunk

    return z_rag


def write_csr_to_zarr(csr_mat, z_arr, start_row=None, end_row=None, ld_boundaries=None, purge_data=False):
    """
    Write from Scipy's csr matrix to Zarr array
    :param csr_mat: A scipy compressed sparse row matrix `csr_matrix`
    :param z_arr: A ragged zarr array with the same row dimension as the csr matrix
    :param start_row: The start row
    :param end_row: The end row
    :param purge_data: If `True`, delete the data that was written to Zarr from `csr_mat`
    :param ld_boundaries: If provided, we'd only write the elements within the provided boundaries.
    """

    if start_row is None:
        start_row = 0

    if end_row is None:
        end_row = csr_mat.shape[0]
    else:
        end_row = min(end_row, csr_mat.shape[0])

    ld_rows = []
    for i in range(start_row, end_row):
        if ld_boundaries is None:
            ld_rows.append(np.nan_to_num(csr_mat[i, :].data))
        else:
            ld_rows.append(np.nan_to_num(csr_mat[i, ld_boundaries[0, i]:ld_boundaries[1, i]].data))

    z_arr.oindex[np.arange(start_row, end_row)] = np.array(ld_rows + [None], dtype=object)[:-1]

    if purge_data:
        # Delete data from csr matrix:
        csr_mat.data[csr_mat.indptr[start_row]:csr_mat.indptr[end_row - 1]] = 0.
        csr_mat.eliminate_zeros()


def from_plink_ld_table_to_zarr_chunked(ld_file, dir_store, ld_boundaries, snps):
    """
    Transform a PLINK LD table to Zarr ragged array.
    PLINK LD tables are of the format:
    CHR_A   BP_A   SNP_A  CHR_B   BP_B  SNP_B  R
    This function deploys a chunked implementation so it only requires
    modest memory.

    :param dir_store: A path to the new Zarr store
    :param ld_file: A path to the plink LD table
    :param ld_boundaries: LD boundaries matrix
    :param snps: A list of SNPs
    """

    # Preliminaries:

    # Estimate row chunk-size for the Zarr array:
    rows, avg_ncols = len(snps), int((ld_boundaries[1, :] - ld_boundaries[0, :]).mean())
    chunks = estimate_row_chunk_size(rows, avg_ncols)

    # Create a ragged Zarr array:
    z_arr = zarr.open(dir_store,
                      mode='w',
                      shape=rows,
                      chunks=chunks[:1],
                      dtype=object,
                      object_codec=VLenArray(float))

    row_chunk_size = z_arr.chunks[0]

    # Create a chunked iterator with pandas:
    # Chunk size will correspond to the average chunk size for the Zarr array:
    ld_chunks = pd.read_csv(ld_file,
                            delim_whitespace=True,
                            usecols=['SNP_A', 'SNP_B', 'R'],
                            engine='c',
                            chunksize=row_chunk_size*avg_ncols // 2)

    # Create a ragged Zarr array:
    z_arr = zarr.open(dir_store,
                      mode='w',
                      shape=rows,
                      chunks=(row_chunk_size,),
                      dtype=object,
                      object_codec=VLenArray(float))

    # Create a dictionary mapping SNPs to their indices:
    snp_dict = dict(zip(snps, np.arange(len(snps))))

    # The sparse matrix will help us convert from triangular
    # sparse matrix to square sparse matrix:
    from scipy.sparse import csr_matrix

    sp_mat = None

    curr_chunk = 0

    # For each chunk in the LD file:
    for ld_chunk in ld_chunks:

        # Create an indexed LD chunk:
        ld_chunk['index_A'] = ld_chunk['SNP_A'].map(snp_dict)
        ld_chunk['index_B'] = ld_chunk['SNP_B'].map(snp_dict)

        ld_chunk['R'].values[ld_chunk['R'].values == 0.] = np.nan

        # Create a compressed sparse row matrix:
        chunk_mat = csr_matrix((ld_chunk['R'].values,
                               (ld_chunk['index_A'].values, ld_chunk['index_B'].values)),
                               shape=(rows, rows))

        if sp_mat is None:
            sp_mat = chunk_mat + chunk_mat.T
            sp_mat.setdiag(1.)
        else:
            sp_mat = sp_mat + (chunk_mat + chunk_mat.T)

        # The chunk of the snp of largest index:
        max_index_chunk = ld_chunk['index_A'].max() // row_chunk_size

        if max_index_chunk > curr_chunk:
            write_csr_to_zarr(sp_mat, z_arr,
                              start_row=curr_chunk*row_chunk_size,
                              end_row=max_index_chunk*row_chunk_size,
                              ld_boundaries=ld_boundaries,
                              purge_data=True)
            curr_chunk = max_index_chunk

    write_csr_to_zarr(sp_mat, z_arr,
                      start_row=curr_chunk * row_chunk_size,
                      ld_boundaries=ld_boundaries,
                      purge_data=True)

    return z_arr


def from_plink_ld_table_to_zarr(ld_file, dir_store, ld_boundaries=None, snps=None):
    """
    Transform a PLINK LD table to Zarr ragged array.
    PLINK LD tables are of the format:
    CHR_A   BP_A   SNP_A  CHR_B   BP_B  SNP_B  R
    :param dir_store: A path to the new Zarr store
    :param ld_file: A path to the plink LD table
    :param ld_boundaries: LD boundaries matrix
    :param snps: A list of SNPs
    """

    ld_df = pd.read_csv(ld_file,
                        delim_whitespace=True,
                        usecols=['BP_A', 'SNP_A', 'BP_B', 'SNP_B', 'R'],
                        engine='c')

    # Assume that PLINK's table is already sorted by BP_A, BP_B:
    a_snps_slice = generate_slice_dictionary(ld_df.SNP_A.values)
    r_a = ld_df.R.values

    ld_sort_b = ld_df.sort_values(['BP_B', 'BP_A'])
    b_snps_slice = generate_slice_dictionary(ld_sort_b.SNP_B.values)
    r_b = ld_sort_b.R.values

    if snps is None:
        snp_df = pd.DataFrame(np.concatenate([ld_df[['SNP_A', 'BP_A']].drop_duplicates().values,
                                              ld_df[['SNP_B', 'BP_B']].drop_duplicates().values]),
                              columns=['SNP', 'BP'])
        snps = snp_df.drop_duplicates().sort_values('BP')['SNP'].values

    if ld_boundaries is None:
        before_bound = np.repeat([None], len(snps))
        after_bound = np.repeat([None], len(snps))
    else:
        before_bound = ld_boundaries[0, :] - np.arange(ld_boundaries.shape[1])
        after_bound = ld_boundaries[1, :] - np.arange(ld_boundaries.shape[1]) - 1

    ld_array = []
    avg_ncol = 0

    for i, snp in enumerate(snps):

        try:
            if before_bound[i] < 0:
                before = r_b[b_snps_slice[snp]][before_bound[i]:]
            else:
                before = []
        except KeyError:
            before = []

        try:
            after = r_a[a_snps_slice[snp]][:after_bound[i]]
        except KeyError:
            after = []

        ld_array.append(np.concatenate([before, [1.], after]))

        avg_ncol += (len(ld_array[-1]) - avg_ncol) / (i + 1)

    n_chunks = estimate_row_chunk_size(len(ld_array), int(avg_ncol))

    z_arr = zarr.open(dir_store,
                      mode='w',
                      shape=len(ld_array),
                      chunks=n_chunks[:1],
                      dtype=object,
                      object_codec=VLenArray(float))

    z_arr[:] = np.array(ld_array, dtype=object)

    return z_arr


def clump_snps(ldm, stat, rsq_threshold=.9, extract=True):
    """
    This function takes an LDMatrix object and clumps the SNPs based
    on the `stat` vector (usually p-value) and the provided r-squared threshold.
    If two SNPs have an r-squared greater than the threshold,
    the SNP with the higher `stat` value is excluded.
    :param ldm: LDMatrix object
    :param stat: A vector of statistics (e.g. p-values) for the SNPs
    :param rsq_threshold: The r^2 threshold to use for filtering
    :param extract: If True, return remaining SNPs. If False, return removed SNPs.
    :return: A list of SNPs passing the specified filter
    """

    snps = ldm.snps
    ld_bounds = ldm.ld_boundaries
    remove_snps = set()

    for idx, ld in enumerate(ldm):

        if snps[idx] in remove_snps:
            continue

        rsq = np.array(ld)**2

        for s_idx in np.where(rsq > rsq_threshold)[0]:
            real_idx = s_idx + ld_bounds[0, idx]
            if idx == real_idx or snps[real_idx] in remove_snps:
                continue

            if stat[idx] < stat[real_idx]:
                remove_snps.add(snps[real_idx])
            else:
                remove_snps.add(snps[idx])

    if extract:
        return list(set(snps) - remove_snps)
    else:
        return list(remove_snps)


def shrink_ld_matrix(arr,
                     cm_pos,
                     maf_var,
                     genmap_Ne,
                     genmap_sample_size,
                     shrinkage_cutoff=1e-3,
                     phased_haplotype=False,
                     ld_boundaries=None):
    """
    Shrink the entries of the LD matrix using the shrinkage estimator
    described in Lloyd-Jones (2019) and Wen and Stephens (2010). The estimator
    is also implemented in the RSS software by Xiang Zhu:

    https://github.com/stephenslab/rss/blob/master/misc/get_corr.R

    :param arr: The Zarr array containing the original LD matrix.
    :param cm_pos: The position of each variant in the LD matrix in centi Morgan.
    :param maf_var: A vector of the variance in minor allele frequency (MAF) for each SNP in the LD matrix. Should be
    equivalent to 2*pj*(1. - pj), where pj is the MAF of SNP j.
    :param genmap_Ne: The effective population size for the genetic map.
    :param genmap_sample_size: The sample size used to estimate the genetic map.
    :param shrinkage_cutoff: The cutoff value below which we assume that the shrinkage factor is zero.
    :param phased_haplotype: A flag indicating whether the LD was calculated from phased haplotypes.
    :param ld_boundaries: The LD boundaries to use when shrinking the LD matrix.
    """

    if ld_boundaries is None:
        ld_boundaries = np.array([np.repeat(None, arr.shape[0]), np.repeat(None, arr.shape[0])])

    # The multiplicative term for the shrinkage factor
    # The shrinkage factor is 4 * Ne * (rho_ij/100) / (2*m)
    # where Ne is the effective population size and m is the sample size
    # for the genetic map and rho_ij is the distance between SNPs i and j
    # in centi Morgan.
    # Therefore, the multiplicative term that we need to apply
    # to the distance between SNPs is: 4*Ne/(200*m), which is equivalent to 0.02*Ne/m
    # See also: https://github.com/stephenslab/rss/blob/master/misc/get_corr.R
    # and Wen and Stephens (2010)

    mult_term = .02*genmap_Ne / genmap_sample_size

    def harmonic_series_sum(n):
        """
        A utility function to compute the sum of the harmonic series
        found in Equation 2.8 in Wen and Stephens (2010)
        Acknowledgement: https://stackoverflow.com/a/27683292
        """
        from scipy.special import digamma
        from numpy import euler_gamma

        return digamma(n + 1) + euler_gamma

    # Compute theta according to Eq. 2.8 in Wen and Stephens (2010)

    h_sum = harmonic_series_sum(2*genmap_sample_size - 1)  # The sum of the harmonic series in Eq. 2.8
    theta = (1. / h_sum) / (2. * genmap_sample_size + 1. / h_sum)  # The theta parameter (related to mutation)
    theta_factor = (1. - theta)**2  # The theta factor that we'll multiply all elements of the covariance matrix with
    theta_diag_factor = .5 * theta * (1. - .5 * theta)  # The theta factor for the diagonal elements

    # Phased haplotype/unphased genotype multiplicative factor:
    # Wen and Stephens (2010), Section 2.4
    phased_mult = [.5, 1.][phased_haplotype]

    # We need to turn the correlation matrix into a covariance matrix to
    # apply the shrinkage factor. For this, we have to multiply each row
    # by the product of standard deviations:
    maf_sd = np.sqrt(maf_var)

    # According to Eqs. 2.6 and 2.7 in Wen and Stephens (2010), the shrunk standard deviation should be:
    shrunk_sd = np.sqrt(theta_factor*maf_var*phased_mult + theta_diag_factor)

    def update_prev_chunk(j):
        """
        A utility function to update the LD matrix chunk by chunk for optimal speed and efficiency.
        """
        chunk_start = (j - 1) - (j - 1) % chunk_size
        chunk_end = min(chunk_start + chunk_size, arr.shape[0])
        arr[chunk_start:chunk_end] = chunk

    chunk_size = arr.chunks[0]
    chunk = None

    for j in range(arr.shape[0]):

        if j % chunk_size == 0:
            if j > 0:
                update_prev_chunk(j)

            chunk = arr[j: j + chunk_size]

        start, end = ld_boundaries[:, j]

        # Compute the shrinkage factor the entries in row j
        shrink_factor = np.exp(-mult_term * np.abs(cm_pos - cm_pos[j])[start: end])
        # Set any shrinkage factor below the cutoff value to zero:
        shrink_factor[shrink_factor < shrinkage_cutoff] = 0.

        # The factor to convert the entries of the correlation matrix into corresponding covariances:
        to_cov_factor = phased_mult*maf_sd[j]*maf_sd[start: end]

        # Compute the theta multiplicative factor following Eq. 2.6 in Wen and Stephens (2010)
        shrink_factor *= theta_factor

        # Compute the new denominator for the Pearson correlation:
        # The shrunk standard deviation of SNP j multiplied by the shrunk standard deviations of each neighbor:
        shrunk_sd_prod = shrunk_sd[j]*shrunk_sd[start: end]

        # Shrink the entries of the LD matrix:
        try:
            shrunk_corr = chunk[j % chunk_size]*to_cov_factor*shrink_factor / shrunk_sd_prod
            shrunk_corr[j - start] = 1.
            chunk[j % chunk_size] = shrunk_corr
        except ValueError:
            raise ValueError(f'Failed to apply shrinkage to row number: {j}')

    update_prev_chunk(j + 1)

    return arr


def sparsify_ld_matrix(arr, bounds):
    """
    A utility function to sparsify chunked LD matrices
    :param arr: the LD matrix
    :param bounds: an 2xM array of start and end position for each row
    :return: A sparsified array of the same format
    """

    def update_prev_chunk(j):
        chunk_start = (j - 1) - (j - 1) % chunk_size
        chunk_end = chunk_start + chunk_size
        arr[chunk_start:chunk_end] = chunk

    chunk_size = arr.chunks[0]
    chunk = None

    for j in range(bounds.shape[1]):
        if j % chunk_size == 0:
            if j > 0:
                update_prev_chunk(j)

            chunk = arr[j: j + chunk_size]

        chunk[j % chunk_size, :bounds[0, j]] = 0
        chunk[j % chunk_size, bounds[1, j]:] = 0

    update_prev_chunk(j + 1)

    return arr


def rechunk_zarr(arr, target_chunks, target_store, intermediate_store, **kwargs):
    """
    Rechunk a Zarr matrix using utilities from `rechunker`.
    """

    if osp.isdir(target_store):
        try:
            z = zarr.open(target_store)
            z.store.rmdir()
        except Exception as e:
            raise e

    from rechunker import rechunk

    rechunked = rechunk(arr,
                        target_chunks=target_chunks,
                        target_store=target_store,
                        temp_store=intermediate_store,
                        max_mem="128MiB",
                        **kwargs)

    try:
        rechunked.execute()
    except Exception as e:
        raise e

    # Delete the older/intermediate stores:
    delete_ld_store(zarr.open(intermediate_store))
    delete_ld_store(arr)

    return zarr.open(target_store)


def optimize_chunks_for_memory(chunked_array, cpus=None, max_mem=None):
    """
    Determine optimal chunks that fit in max_mem. Max_mem should be numerical in GiB
    Modified from: Sergio Hleap
    """

    import psutil
    import dask.array as da

    if cpus is None:
        cpus = psutil.cpu_count()

    if max_mem is None:
        max_mem = psutil.virtual_memory().available / (1024.0 ** 3)

    chunk_mem = max_mem / cpus
    chunks = da.core.normalize_chunks(f"{chunk_mem}GiB", shape=chunked_array.shape, dtype=chunked_array.dtype)

    return chunked_array.chunk(chunks)


def estimate_row_chunk_size(rows, cols, dtype=np.float64, chunk_size=128):
    """
    Estimate the chunk size for ragged arrays, given the number of rows, columns, and data type.

    :param rows: Number of rows.
    :param cols: Number of columns. If a ragged array, provide average size of arrays
    :param dtype: Data type
    :param chunk_size: chunk size in MB
    """

    matrix_size = rows * cols * np.dtype(dtype).itemsize / 1024 ** 2
    n_chunks = matrix_size // chunk_size

    if n_chunks < 1:
        return None, None
    else:
        return int(rows / n_chunks), None


def dense_zarr_array_to_ragged(z,
                               dir_store,
                               ld_boundaries,
                               rechunk=True,
                               delete_original=True):
    """
    This function takes a dense chunked Zarr matrix
    and, given an array of window sizes for each row, returns a sparse ragged array matrix.
    This is a utility function that works with `dask` or `xarray` generated Linkage-Disequilibrium (LD) matrices
    and aims to create compact LD matrices that are easier to manipulate and work with.

    :param z: The original LD matrix in Zarr format.
    :param dir_store: The path to the new store where the sparse LD matrix will be stored.
    :param ld_boundaries: The LD boundaries or window around each SNP. This is a 2xM array where
    the first row contains the start and the second row contains the end of each window.
    :param rechunk: If True, re-chunk the ragged array for optimized read/write performance.
    :param delete_original: Delete the original store after creating the ragged array.

    """

    avg_ncol = int((ld_boundaries[1, :] - ld_boundaries[0, :]).mean())

    if rechunk:
        n_chunks = estimate_row_chunk_size(z.shape[0], avg_ncol)
    else:
        n_chunks = z.chunks

    if avg_ncol == z.shape[0]:
        z_rag = zarr.open(dir_store,
                          mode='w',
                          shape=z.shape,
                          chunks=n_chunks,
                          dtype=float)
    else:
        z_rag = zarr.open(dir_store,
                          mode='w',
                          shape=z.shape[0],
                          chunks=n_chunks[:1],
                          dtype=object,
                          object_codec=VLenArray(float))

    chunk_size = z.chunks[0]

    for i in range(int(np.ceil(z.shape[0] / chunk_size))):

        start = i * chunk_size
        end = min((i + 1) * chunk_size, z.shape[0])

        z_chunk = z[start: end]

        z_rag_rows = []

        for j in range(start, end):
            z_rag_rows.append(
                z_chunk[j - start][ld_boundaries[0, j]:ld_boundaries[1, j]]
            )

        if avg_ncol == z.shape[0]:
            z_rag.oindex[np.arange(start, end)] = np.array(z_rag_rows)
        else:
            z_rag.oindex[np.arange(start, end)] = np.array(z_rag_rows + [None], dtype=object)[:-1]

    if delete_original:
        delete_ld_store(z)

    return z_rag


def filter_zarr_array(z,
                      dir_store,
                      extract_snps,
                      ld_boundaries,
                      rechunk=True,
                      delete_original=False):
    """
    This function takes a chunked Zarr matrix (dense or sparse LD matrix)
    and, given a list of SNPs to extract, returns a filtered ragged array matrix.

    TODO: Optimize this for large chromosomes/LD matrices!

    :param z: the original Zarr matrix (implementation assumes 2D matrix)
    :param dir_store: The path to the new Zarr matrix store
    :param extract_snps: A list or vector of SNP IDs to keep.
    :param ld_boundaries: The LD boundaries or window around each SNP. This is a 2xM array where
    the first row contains the start and the second row contains the end of each window.
    :param rechunk: If True, re-chunk the filtered array for optimized read/write performance.
    :param delete_original: If True, delete the original store after transformation.
    """

    idx_map = pd.DataFrame({'SNP': extract_snps}).reset_index().merge(
        pd.DataFrame({'SNP': z.attrs['SNP']}).reset_index(),
        on='SNP',
        suffixes=('_y', '_x')
    )
    idx_map['chunk_x'] = (idx_map['index_x'] // z.chunks[0]).astype(int)
    n_rows = len(extract_snps)

    idx_map['chunk_x'] = (idx_map['index_x'] // z.chunks[0]).astype(int)

    orig_bounds = np.array(z.attrs['LD boundaries'])

    avg_ncol = int((ld_boundaries[1, :] - ld_boundaries[0, :]).mean())

    if rechunk:
        n_chunks = estimate_row_chunk_size(n_rows, avg_ncol)
    else:
        n_chunks = z.chunks

    if avg_ncol == n_rows:
        z_rag = zarr.open(dir_store,
                          mode='w',
                          shape=(n_rows, n_rows),
                          chunks=n_chunks,
                          dtype=float)
    else:
        z_rag = zarr.open(dir_store,
                          mode='w',
                          shape=n_rows,
                          chunks=n_chunks[:1],
                          dtype=object,
                          object_codec=VLenArray(float))

    idx_x = idx_map['index_x'].values
    chunk_size = z.chunks[0]

    for i in range(int(np.ceil(z.shape[0] / chunk_size))):

        start = i * chunk_size
        end = min((i + 1) * chunk_size, z.shape[0])

        z_chunk = z[start: end]

        z_rag_index = []
        z_rag_rows = []

        for _, (k, _, j, _) in idx_map.loc[idx_map['chunk_x'] == i].iterrows():

            z_rag_index.append(k)

            # Find the index of SNPs in the original LD matrix that
            # remain after matching with the `keep_snps` variable.
            orig_idx = idx_x[(orig_bounds[0, j] <= idx_x) & (idx_x < orig_bounds[1, j])] - orig_bounds[0, j]
            row_val = z_chunk[j - start][orig_idx]

            z_rag_rows.append(row_val)

        if len(z_rag_index) == 0:
            continue

        if avg_ncol == n_rows:
            z_rag.oindex[z_rag_index] = np.array(z_rag_rows)
        else:
            z_rag.oindex[z_rag_index] = np.array(z_rag_rows + [None], dtype=object)[:-1]

    z_rag.attrs.update(z.attrs.asdict())

    # Update the attributes associated with the new matrix:
    z_rag.attrs['LD boundaries'] = ld_boundaries.tolist()

    try:
        z_rag.attrs['SNP'] = list(extract_snps)
    except Exception:
        pass

    try:
        z_rag.attrs['BP'] = list(map(int, np.array(z.attrs['BP'])[idx_x]))
    except Exception:
        pass

    try:
        z_rag.attrs['cM'] = list(map(float, np.array(z.attrs['cM'])[idx_x]))
    except Exception:
        pass

    try:
        z_rag.attrs['MAF'] = list(map(float, np.array(z.attrs['MAF'])[idx_x]))
    except Exception:
        pass

    try:
        z_rag.attrs['A1'] = list(np.array(z.attrs['A1'])[idx_x])
    except Exception:
        pass

    try:
        z_rag.attrs['A2'] = list(np.array(z.attrs['A2'])[idx_x])
    except Exception:
        pass

    try:
        z_rag.attrs['LDScore'] = list(map(float, np.array(z.attrs['LDScore'])[idx_x]))
    except Exception:
        pass

    if delete_original:
        delete_ld_store(z)

    return z_rag


def compute_ld_plink1p9(genotype_matrix,
                        ld_boundaries,
                        output_dir,
                        temp_dir='temp'):

    from magenpy.utils.executors import plink1Executor
    from magenpy.GenotypeMatrix import plinkBEDGenotypeMatrix

    assert isinstance(genotype_matrix, plinkBEDGenotypeMatrix)

    plink1 = plink1Executor()

    keep_file = osp.join(temp_dir, 'samples.keep')
    keep_table = genotype_matrix.sample_table.get_individual_table()
    keep_table.to_csv(keep_file, index=False, header=False, sep="\t")

    snp_keepfile = osp.join(temp_dir, 'variants.keep')
    pd.DataFrame({'SNP': genotype_matrix.snps}).to_csv(
        snp_keepfile, index=False, header=False
    )

    plink_output = osp.join(temp_dir, f'chr_{str(genotype_matrix.chromosome)}')

    # Set the window sizes in various units:
    # (1) Number of neighboring SNPs:
    window_size = (ld_boundaries - np.arange(genotype_matrix.m)).max() + 10

    # (2) Kilobases:

    positional_bounds = np.clip(np.array([ld_boundaries[0, :] - 1, ld_boundaries[1, :]]),
                                a_min=0, a_max=ld_boundaries.shape[1] - 1)

    kb_pos = .001*genotype_matrix.bp_pos
    kb_bounds = kb_pos[positional_bounds]
    kb_window_size = (kb_bounds - kb_pos).max() + .01

    # (3) centi Morgan:
    try:
        cm_pos = genotype_matrix.cm_pos
        cm_bounds = genotype_matrix.cm_pos[positional_bounds]
        cm_window_size = (cm_bounds - cm_pos).max() + .01
    except Exception:
        cm_window_size = None

    cmd = [
        f"--bfile {genotype_matrix.bed_file.replace('.bed', '')}",
        f"--keep {keep_file}",
        f"--extract {snp_keepfile}",
        "--keep-allele-order",
        f"--out {plink_output}",
        "--r gz",
        f"--ld-window {window_size}",
        f"--ld-window-kb {kb_window_size}"
    ]

    if cm_window_size is not None:
        cmd.append(f"--ld-window-cm {cm_window_size}")

    plink1.execute(cmd)

    # Convert from PLINK LD files to Zarr:
    fin_ld_store = osp.join(output_dir, 'ld', 'chr_' +
                            str(genotype_matrix.chromosome))

    z_ld_mat = from_plink_ld_table_to_zarr_chunked(f"{plink_output}.ld.gz",
                                                   fin_ld_store,
                                                   ld_boundaries,
                                                   genotype_matrix.snps)

    return z_ld_mat


def compute_ld_xarray(genotype_matrix,
                      ld_boundaries,
                      output_dir,
                      temp_dir='temp'):
    """
    Compute the Linkage Disequilibrium matrix or snp-by-snp
    correlation matrix assuming that the genotypes are represented
    by `xarray` or `dask`-like matrix. This function computes the
    entire X'X/N and stores the result in Zarr arrays.
    To create sparse matrices out of this, consult the
    LD estimators and their implementations.

    NOTE: We don't recommend using this for large-scale genotype matrices.
    Use `compute_ld_plink` instead if you have plink installed on your system.

    :param genotype_matrix: An `xarrayGenotypeMatrix` object
    :param temp_dir: A temporary directory where to store intermediate results.

    """

    from magenpy.GenotypeMatrix import xarrayGenotypeMatrix
    assert isinstance(genotype_matrix, xarrayGenotypeMatrix)

    g_data = genotype_matrix.xr_mat

    # Re-chunk the array
    g_data = g_data.chunk((min(1024, g_data.shape[0]), min(1024, g_data.shape[1])))

    from ..transforms.genotype import standardize
    import dask.array as da

    # Standardize the genotype matrix and fill missing data with zeros:
    g_mat = standardize(g_data).data

    # Compute the LD matrix:
    ld_mat = (da.dot(g_mat.T, g_mat) / genotype_matrix.sample_size).astype(np.float64)
    ld_mat.to_zarr(temp_dir, overwrite=True)

    z_ld_mat = zarr.open(temp_dir)
    z_ld_mat = rechunk_zarr(z_ld_mat,
                            ld_mat.rechunk({0: 'auto', 1: None}).chunksize,
                            temp_dir + '_rechunked',
                            temp_dir + '_intermediate')

    fin_ld_store = osp.join(output_dir, 'ld', 'chr_' +
                            str(genotype_matrix.chromosome))

    # If the matrix is sparse/thresholded, then convert to a ragged zarr array:
    if (ld_boundaries[1, :] - ld_boundaries[0, :]).min() < genotype_matrix.n_snps:

        z_ld_mat = dense_zarr_array_to_ragged(z_ld_mat,
                                              fin_ld_store,
                                              ld_boundaries)

    else:
        z_ld_mat = move_ld_store(z_ld_mat, fin_ld_store)

    return z_ld_mat



from typing import Union
import os.path as osp
import pandas as pd
import numpy as np
import warnings
import magenpy as mgp


def inflation_factor(gdl: Union[mgp.GWADataLoader, None] = None,
                     sumstats: Union[mgp.SumstatsTable, None] = None,
                     chisq=None):
    """
    Compute the genomic control (GC) inflation factor (also known as lambda)
    from GWAS summary statistics.

    The inflation factor can be used to detect and correct inflation in the test statistics.

    :param chisq: An array of chi-squared statistics to compute the inflation factor from.
    :param gdl: A `GWADataLoader` object, with summary statistics initilized into the `sumstats_table` property.
    :param sumstats: A `SumstatsTable` object, with the GWAS summary statistics load and initialized.
    """

    assert chisq is not None or gdl is not None or sumstats is not None

    if chisq is None:
        if gdl is not None:
            chisq = np.concatenate([ss.get_chisq_statistic() for ss in gdl.sumstats_table.values()])
        else:
            chisq = sumstats.get_chisq_statistic()

    from scipy.stats import chi2

    return np.median(chisq) / chi2.median(1)


def perform_gwa_plink2(genotype_matrix,
                       temp_dir='temp',
                       standardize_phenotype=True,
                       include_covariates=True):

    from magenpy.GenotypeMatrix import plinkBEDGenotypeMatrix
    from magenpy.utils.executors import plink2Executor

    assert isinstance(genotype_matrix, plinkBEDGenotypeMatrix)

    plink2 = plink2Executor()

    s_table = genotype_matrix.sample_table

    if s_table.phenotype_likelihood is None:
        warnings.warn("The phenotype likelihood is not specified! "
                      "Assuming that the phenotype is continuous...")

    # Output phenotype table:
    phe_fname = osp.join(temp_dir, "pheno.txt")
    phe_table = s_table.get_phenotype_table()
    if s_table.phenotype_likelihood == 'binomial':
        phe_table['phenotype'] += 1
    phe_table.to_csv(phe_fname, sep="\t", index=False, header=False)

    # Process covariates:
    if include_covariates and s_table.covariates is not None:
        covar_fname = osp.join(temp_dir, "covar.txt")
        covar = s_table.get_covariates_table()
        covar.to_csv(covar_fname, sep="\t", index=False, header=False)
        covar_modifier = ''
    else:
        covar_fname = None
        covar_modifier = ' allow-no-covars'

    # Determine regression type based on phenotype likelihood:
    plink_reg_type = ['linear', 'logistic'][s_table.phenotype_likelihood == 'binomial']

    # Output subset of SNPs to perform association tests on:
    snp_keepfile = osp.join(temp_dir, f"variants.keep")
    pd.DataFrame({'SNP': genotype_matrix.snps}).to_csv(
        snp_keepfile, index=False, header=False
    )

    # Output file:
    plink_output = osp.join(temp_dir, f"output")

    cmd = [
        f"--bfile {genotype_matrix.bed_file}",
        f"--extract {snp_keepfile}",
        f"--{plink_reg_type} hide-covar{covar_modifier} cols=chrom,pos,alt1,ref,a1freq,nobs,beta,se,tz,p",
        f"--pheno {phe_fname}",
        f"--out {plink_output}"
    ]

    if standardize_phenotype and plink_reg_type == 'linear':
        cmd.append('--variance-standardize')

    if covar_fname is not None:
        cmd.append(f'--covar {covar_fname}')

    plink2.execute(cmd)

    output_fname = plink_output + f".PHENO1.glm.{plink_reg_type}"

    if not osp.isfile(output_fname):
        if plink_reg_type == 'logistic' and osp.isfile(output_fname + ".hybrid"):
            output_fname += ".hybrid"
        else:
            raise FileNotFoundError

    # Read the summary statistics file from plink:
    ss_table = mgp.SumstatsTable.from_file(output_fname, sumstats_format='plink')
    # Make sure that the effect allele is encoded properly:
    ss_table.match(genotype_matrix.snp_table, correct_flips=True)

    return ss_table


def perform_gwa_plink1p9(genotype_matrix, standardize_phenotype=False):
    """
    TODO: Add support for performing association testing with plink1.9
    """

    #assert isinstance(genotype_matrix, plinkBEDGenotypeMatrix)
    raise NotImplementedError


def perform_gwa_xarray(genotype_matrix,
                       standardize_genotype=True,
                       standardize_phenotype=True):

    from magenpy.GenotypeMatrix import xarrayGenotypeMatrix
    assert isinstance(genotype_matrix, xarrayGenotypeMatrix)

    if genotype_matrix.sample_table.phenotype_likelihood is None:
        warnings.warn("The phenotype likelihood is not specified! "
                      "Assuming that the phenotype is continuous...")
    elif genotype_matrix.sample_table.phenotype_likelihood == 'binomial':
        raise Exception("The xarray backend does not support performing association "
                        "testing on binary (case-control) phenotypes!")

    sumstats_table = genotype_matrix.get_snp_table(
        ['CHR', 'SNP', 'POS', 'A1', 'A2', 'N', 'MAF']
    )

    phenotype = genotype_matrix.sample_table.phenotype

    if standardize_phenotype:
        from ..transforms.phenotype import standardize
        phenotype = standardize(phenotype)

    sigma_sq_y = np.var(phenotype)

    if standardize_genotype:

        from ..transforms.genotype import standardize

        sumstats_table['BETA'] = np.dot(standardize(genotype_matrix.xr_mat).T, phenotype) / sumstats_table['N'].values
        sumstats_table['SE'] = np.sqrt(sigma_sq_y / sumstats_table['N'].values)
    else:

        sumstats_table['BETA'] = (
            np.dot(genotype_matrix.xr_mat.fillna(sumstats_table['MAF'].values).T, phenotype) /
            sumstats_table['N'].values * genotype_matrix.maf_var
        )

        sumstats_table['SE'] = np.sqrt(sigma_sq_y / (sumstats_table['N'].values * genotype_matrix.maf_var))

    ss_table = mgp.SumstatsTable(sumstats_table)
    _, _ = ss_table.z_score, ss_table.pval

    return ss_table




from typing import Union
import tempfile
import pandas as pd
import numpy as np
from .parsers.plink_parsers import parse_fam_file, parse_bim_file
from .SampleTable import SampleTable


class GenotypeMatrix(object):

    def __init__(self,
                 sample_table: Union[pd.DataFrame, SampleTable, None] = None,
                 snp_table: Union[pd.DataFrame, None] = None,
                 temp_dir: str = 'temp',
                 **kwargs):

        self.sample_table: Union[pd.DataFrame, SampleTable, None] = None
        self.snp_table: Union[pd.DataFrame, None] = snp_table

        if sample_table is not None:
            self.set_sample_table(sample_table)

        from .utils.system_utils import makedir
        makedir(temp_dir)
        self.temp_dir = temp_dir
        self.cleanup_dir_list = []  # Directories to clean up after execution.

    @classmethod
    def from_file(cls, file_path, temp_dir='temp'):
        """
        Read and parse the genotype matrix information from file.
        """
        raise NotImplementedError

    @property
    def shape(self):
        """
        The shape of the genotype matrix. Rows correspond to the
        number of samples and columns to the number of SNPs.
        """
        return self.n, self.m

    @property
    def n(self):
        """
        The sample size, see also `.sample_size()`
        """
        return self.sample_table.n

    @property
    def sample_size(self):
        """
        The sample size of the genotype matrix. See also `.n()`.
        """
        return self.n

    @property
    def samples(self):
        """
        Obtain a vector of sample IDs.
        """
        return self.sample_table.iid

    @property
    def m(self):
        """
        The number of SNPs, see also `n_snps`
        """
        if self.snp_table is not None:
            return len(self.snp_table)

    @property
    def n_snps(self):
        return self.m

    @property
    def chromosome(self):
        """
        If the genotype matrix is comprised of a single chromosome, return the chromosome number.
        """
        chrom = self.chromosomes
        if chrom is not None:
            if len(chrom) == 1:
                return chrom[0]

    @property
    def chromosomes(self):
        """
        Return the unique set of chromosomes comprising the genotype matrix.
        """
        chrom = self.get_snp_attribute('CHR')
        if chrom is not None:
            return np.unique(chrom)

    @property
    def snps(self):
        """
        Return the SNP IDs.
        """
        return self.get_snp_attribute('SNP')

    @property
    def bp_pos(self):
        """
        The position for the genetic variants in base pairs.
        """
        return self.get_snp_attribute('POS')

    @property
    def cm_pos(self):
        """
        The position for the genetic variants in Centi Morgan.
        """
        cm = self.get_snp_attribute('cM')
        if len(set(cm)) == 1:
            raise KeyError("Genetic distance in centi Morgan (cM) is not "
                           "set in the genotype file!")
        return cm

    @property
    def a1(self):
        """
        Return the effect allele `A1`. See also `.alt_allele()`, `.effect_allele()`.
        """
        return self.get_snp_attribute('A1')

    @property
    def a2(self):
        """
        Return the reference allele `A2`. See also `.ref_allele()`.
        """
        return self.get_snp_attribute('A2')

    @property
    def ref_allele(self):
        """
        Return the reference allele `A2`. See also `.a2()`.
        """
        return self.a2

    @property
    def alt_allele(self):
        """
        Return the alternative (i.e. effect) allele `A1`. See also `.a1()`, `.effect_allele()`.
        """
        return self.a1

    @property
    def effect_allele(self):
        """
        Return the effect allele `A1`. See also `.a1()`, `.alt_allele()`.
        """
        return self.a1

    @property
    def n_per_snp(self):
        """
        Sample size per genetic variant (this accounts for missing values).
        """
        n = self.get_snp_attribute('N')
        if n is not None:
            return n
        else:
            self.compute_sample_size_per_snp()
            return self.get_snp_attribute('N')

    @property
    def maf(self):
        """
        Minor allele frequency
        """
        maf = self.get_snp_attribute('MAF')
        if maf is not None:
            return maf
        else:
            self.compute_allele_frequency()
            return self.get_snp_attribute('MAF')

    @property
    def maf_var(self):
        """
        The variance in minor allele frequency.
        """
        return 2. * self.maf * (1. - self.maf)

    def estimate_memory_allocation(self, dtype=np.float32):
        """
        Estimate the size of the genotype matrix in MB
        """
        return self.n * self.m * np.dtype(dtype).itemsize / 1024 ** 2

    def get_snp_table(self, col_subset=None):
        """
        Return the SNP table or a subset of its columns.
        """

        if col_subset is None:
            return self.snp_table.copy()
        else:
            present_cols = list(set(col_subset).intersection(set(self.snp_table.columns)))
            non_present_cols = list(set(col_subset) - set(present_cols))

            if len(present_cols) > 0:
                table = self.snp_table[present_cols].copy()
            else:
                table = pd.DataFrame({c: [] for c in non_present_cols})

            for col in non_present_cols:

                if col == 'MAF':
                    table['MAF'] = self.maf
                elif col == 'MAF_VAR':
                    table['MAF_VAR'] = self.maf_var
                elif col == 'N':
                    table['N'] = self.n_per_snp
                else:
                    raise KeyError(f"Column '{col}' is not available in the SNP table!")

            return table[list(col_subset)]

    def get_snp_attribute(self, attr):
        """
        A utility function to extract a given column from the SNP table.
        """
        if self.snp_table is not None:
            if attr in self.snp_table.columns:
                return self.snp_table[attr].values

    def compute_ld(self, estimator, output_dir, **ld_kwargs):
        """
        Compute the Linkage-Disequilibrium (LD) or SNP-by-SNP correlation matrix
        for the genotype matrix.

        :param estimator: The estimator for the LD matrix. We currently support
        4 different estimators: `sample`, `windowed`, `shrinkage`, and `block`.
        :param output_dir: The output directory where the Zarr array containing the
        entries of the LD matrix will be stored.
        :param ld_kwargs: keyword arguments for the various LD estimators. Consult
        the implementations of `WindowedLD`, `ShrinkageLD`, and `BlockLD` for details.
        """

        from .stats.ld.estimator import SampleLD, WindowedLD, ShrinkageLD, BlockLD

        if estimator == 'sample':
            ld_est = SampleLD(self)
        elif estimator == 'windowed':
            ld_est = WindowedLD(self, **ld_kwargs)
        elif estimator == 'shrinkage':
            ld_est = ShrinkageLD(self, **ld_kwargs)
        elif estimator == 'block':
            ld_est = BlockLD(self, **ld_kwargs)
        else:
            raise KeyError(f"LD estimator {estimator} is not recognized!")

        # Create a temporary directory where we store intermediate results:
        tmp_ld_dir = tempfile.TemporaryDirectory(dir=self.temp_dir, prefix='ld_')
        self.cleanup_dir_list.append(tmp_ld_dir)

        return ld_est.compute(output_dir, temp_dir=tmp_ld_dir.name)

    def set_sample_table(self, sample_table):
        """
        A convenience method set the sample table for genotype matrix.
        This may be useful for syncing sample tables across different Genotype matrices
        corresponding to different chromosomes or genomic regions.
        """

        if isinstance(sample_table, SampleTable):
            self.sample_table = sample_table
        elif isinstance(sample_table, pd.DataFrame):
            self.sample_table = SampleTable(sample_table)
        else:
            raise Exception("The sample table is invalid!")

    def filter_snps(self, extract_snps=None, extract_file=None):
        """
        Filter variants from the genotype matrix. User must specify
        either a list of variants to extract or the path to a file
        with the list of variants to extract.

        :param extract_snps: A list (or array) of SNP IDs to keep in the genotype matrix.
        :param extract_file: The path to a file with the list of variants to extract.
        """

        assert extract_snps is not None or extract_file is not None

        if extract_snps is None:
            from .parsers.misc_parsers import read_snp_filter_file
            extract_snps = read_snp_filter_file(extract_file)

        self.snp_table = self.snp_table.merge(pd.DataFrame({'SNP': extract_snps}))

    def filter_by_allele_frequency(self, min_maf=None, min_mac=1):
        """
        Filter variants by minimum minor allele frequency or allele count
        :param min_maf: Minimum minor allele frequency
        :param min_mac: Minimum minor allele count (1 by default)
        """

        if min_mac or min_maf:
            maf = self.maf
            n = self.n_per_snp

        keep_flag = None

        if min_mac:
            mac = (2*maf*n).astype(np.int64)
            keep_flag = (mac >= min_mac) & ((2*n - mac) >= min_mac)

        if min_maf:

            maf_cond = (maf >= min_maf) & (1. - maf >= min_maf)
            if keep_flag is not None:
                keep_flag = keep_flag & maf_cond
            else:
                keep_flag = maf_cond

        if keep_flag is not None:
            self.filter_snps(extract_snps=self.snps[keep_flag])

    def drop_duplicated_snps(self):
        """
        Drop variants with duplicated SNP IDs.
        """

        u_snps, counts = np.unique(self.snps, return_counts=True)
        if len(u_snps) < self.n_snps:
            # Keep only SNPs which occur once in the sequence:
            self.filter_snps(u_snps[counts == 1])

    def filter_samples(self, keep_samples=None, keep_file=None):
        """
        Filter samples from the genotype matrix. User must specify
        either a list of samples to keep or the path to a file
        with the list of samples to keep.

        :param keep_samples: A list (or array) of sample IDs to keep in the genotype matrix.
        :param keep_file: The path to a file with the list of samples to keep.
        """

        self.sample_table.filter_samples(keep_samples=keep_samples, keep_file=keep_file)

    def score(self, beta, standardize_genotype=False):
        """
        Perform linear scoring, i.e. multiply the genotype matrix by the vector of effect sizes, `beta`.

        :param beta: A vector of effect sizes for each variant in the genotype matrix.
        :param standardize_genotype: If True, standardized the genotype matrix when computing the score.
        """
        raise NotImplementedError

    def perform_gwas(self, **gwa_kwargs):
        """
        Perform genome-wide association testing of all variants against the phenotype.

        :param gwa_kwargs: Keyword arguments to pass to the GWA functions. Consult stats.gwa.utils
        for relevant keyword arguments for each backend.
        """
        raise NotImplementedError

    def compute_allele_frequency(self):
        """
        Compute the allele frequency of each variant or SNP in the genotype matrix.
        """
        raise NotImplementedError

    def compute_sample_size_per_snp(self):
        """
        Compute the sample size for each variant in the genotype matrix, accounting for
        potential missing values.
        """
        raise NotImplementedError

    def split_by_chromosome(self):
        """
        Split the genotype matrix by chromosome, so that we would
        have a separate `GenotypeMatrix` objects for each chromosome.
        This method returns a dictionary where the key is the chromosome number
        and the value is an object of `GenotypeMatrix` for that chromosome.
        """

        chromosome = self.chromosome

        if chromosome:
            return {chromosome: self}
        else:
            chrom_tables = self.snp_table.groupby('CHR')
            return {
                c: self.__class__(sample_table=self.sample_table,
                                  snp_table=chrom_tables.get_group(c),
                                  temp_dir=self.temp_dir)
                for c in chrom_tables.groups
            }

    def cleanup(self):
        """
        Clean up all temporary files and directories
        """

        for tmpdir in self.cleanup_dir_list:
            try:
                tmpdir.cleanup()
            except FileNotFoundError:
                continue


class xarrayGenotypeMatrix(GenotypeMatrix):

    def __init__(self, sample_table=None, snp_table=None, temp_dir='temp', xr_mat=None):
        super().__init__(sample_table=sample_table, snp_table=snp_table, temp_dir=temp_dir)

        # xarray matrix object, as defined by pandas-plink:
        self.xr_mat = xr_mat

    @classmethod
    def from_file(cls, file_path, temp_dir='temp'):

        from pandas_plink import read_plink1_bin

        try:
            xr_gt = read_plink1_bin(file_path + ".bed", ref="a0", verbose=False)
        except ValueError:
            xr_gt = read_plink1_bin(file_path, ref="a0", verbose=False)
        except Exception as e:
            raise e

        # Set the sample table:
        sample_table = xr_gt.sample.coords.to_dataset().to_dataframe()
        sample_table.columns = ['FID', 'IID', 'fatherID', 'motherID', 'sex', 'phenotype']
        sample_table.reset_index(inplace=True, drop=True)
        sample_table = sample_table.astype({
            'FID': str,
            'IID': str,
            'fatherID': str,
            'motherID': str,
            'sex': float,
            'phenotype': float
        })

        sample_table['phenotype'] = sample_table['phenotype'].replace({-9.: np.nan})

        # Set the snp table:
        snp_table = xr_gt.variant.coords.to_dataset().to_dataframe()
        snp_table.columns = ['CHR', 'SNP', 'cM', 'POS', 'A1', 'A2']
        snp_table.reset_index(inplace=True, drop=True)
        snp_table = snp_table.astype({
            'CHR': int,
            'SNP': str,
            'cM': float,
            'POS': np.int,
            'A1': str,
            'A2': str
        })

        # Set the index to be the SNP ID:
        xr_gt = xr_gt.set_index(variant='snp')

        g_mat = cls(sample_table=SampleTable(sample_table),
                    snp_table=snp_table,
                    temp_dir=temp_dir,
                    xr_mat=xr_gt)

        return g_mat

    def set_sample_table(self, sample_table):

        super(xarrayGenotypeMatrix, self).set_sample_table(sample_table)

        try:
            if self.n != self.xr_mat.shape[0]:
                self.xr_mat = self.xr_mat.sel(sample=self.samples)
        except AttributeError:
            pass

    def filter_snps(self, extract_snps=None, extract_file=None):

        super(xarrayGenotypeMatrix, self).filter_snps(extract_snps=extract_snps, extract_file=extract_file)
        self.xr_mat = self.xr_mat.sel(variant=self.snps)

    def filter_samples(self, keep_samples=None, keep_file=None):

        super(xarrayGenotypeMatrix, self).filter_samples(keep_samples=keep_samples, keep_file=keep_file)
        self.xr_mat = self.xr_mat.sel(sample=self.samples)

    def score(self, beta, standardize_genotype=False, skip_na=True):
        """
        Perform linear scoring on the genotype matrix.
        :param beta: A vector or matrix of effect sizes for each variant in the genotype matrix.
        :param standardize_genotype: If True, standardize the genotype when computing the polygenic score.
        :param skip_na: If True, skip missing values when computing the polygenic score.
        """

        import dask.array as da

        chunked_beta = da.from_array(beta, chunks=self.xr_mat.data.chunksize[1])

        if standardize_genotype:
            from .stats.transforms.genotype import standardize
            pgs = da.dot(standardize(self.xr_mat).data, chunked_beta).compute()
        else:
            if skip_na:
                pgs = da.dot(da.nan_to_num(self.xr_mat.data), chunked_beta).compute()
            else:
                pgs = da.dot(self.xr_mat.fillna(self.maf).data, chunked_beta).compute()

        return pgs

    def perform_gwas(self, **gwa_kwargs):

        from magenpy.stats.gwa.utils import perform_gwa_xarray
        return perform_gwa_xarray(self, **gwa_kwargs)

    def compute_allele_frequency(self):
        self.snp_table['MAF'] = (self.xr_mat.sum(axis=0) / (2. * self.n_per_snp)).compute().values

    def compute_sample_size_per_snp(self):
        self.snp_table['N'] = self.xr_mat.shape[0] - self.xr_mat.isnull().sum(axis=0).compute().values

    def split_by_chromosome(self):
        split = super(xarrayGenotypeMatrix, self).split_by_chromosome()

        for c, gt in split.items():
            if len(split) > 1:
                gt.xr_mat = self.xr_mat.sel(variant=gt.snps)
            else:
                gt.xr_mat = self.xr_mat

        return split


class plinkBEDGenotypeMatrix(GenotypeMatrix):

    def __init__(self, sample_table=None, snp_table=None, temp_dir='temp', bed_file=None):
        super().__init__(sample_table=sample_table, snp_table=snp_table, temp_dir=temp_dir)

        self.bed_file = bed_file
        if self.bed_file is not None:
            self.bed_file = self.bed_file.replace('.bed', '')

        if self.sample_table is None and self.bed_file:
            self.sample_table = SampleTable(parse_fam_file(self.bed_file))

        if self.snp_table is None and self.bed_file:
            self.snp_table = parse_bim_file(self.bed_file)

    @classmethod
    def from_file(cls, file_path, temp_dir='temp'):

        p_gt = cls(bed_file=file_path, temp_dir=temp_dir)

        return p_gt

    def score(self, beta, standardize_genotype=False):

        from .stats.score.utils import score_plink2

        # Create a temporary directory where we store intermediate results:
        tmp_score_dir = tempfile.TemporaryDirectory(dir=self.temp_dir, prefix='score_')
        self.cleanup_dir_list.append(tmp_score_dir)

        return score_plink2(self, beta, standardize_genotype=standardize_genotype, temp_dir=tmp_score_dir.name)

    def perform_gwas(self, **gwa_kwargs):

        from magenpy.stats.gwa.utils import perform_gwa_plink2

        # Create a temporary directory where we store intermediate results:
        tmp_gwas_dir = tempfile.TemporaryDirectory(dir=self.temp_dir, prefix='gwas_')
        self.cleanup_dir_list.append(tmp_gwas_dir)

        return perform_gwa_plink2(self, temp_dir=tmp_gwas_dir.name, **gwa_kwargs)

    def compute_allele_frequency(self):

        from magenpy.stats.variant.utils import compute_allele_frequency_plink2

        # Create a temporary directory where we store intermediate results:
        tmp_freq_dir = tempfile.TemporaryDirectory(dir=self.temp_dir, prefix='freq_')
        self.cleanup_dir_list.append(tmp_freq_dir)

        self.snp_table['MAF'] = compute_allele_frequency_plink2(self, temp_dir=tmp_freq_dir.name)

    def compute_sample_size_per_snp(self):
        from magenpy.stats.variant.utils import compute_sample_size_per_snp_plink2

        # Create a temporary directory where we store intermediate results:
        tmp_miss_dir = tempfile.TemporaryDirectory(dir=self.temp_dir, prefix='miss_')
        self.cleanup_dir_list.append(tmp_miss_dir)

        self.snp_table['N'] = compute_sample_size_per_snp_plink2(self, temp_dir=tmp_miss_dir.name)

    def split_by_chromosome(self):

        split = super(plinkBEDGenotypeMatrix, self).split_by_chromosome()

        for c, gt in split.items():
            gt.bed_file = self.bed_file

        return split



import os.path as osp
import numpy as np
import pandas as pd


def score_plink2(genotype_matrix,
                 betas,
                 standardize_genotype=False,
                 temp_dir='temp'):

    from magenpy.GenotypeMatrix import plinkBEDGenotypeMatrix
    from magenpy.utils.executors import plink2Executor

    assert isinstance(genotype_matrix, plinkBEDGenotypeMatrix)

    plink2 = plink2Executor()

    try:
        betas_shape = betas.shape[1]
        if betas_shape == 1:
            raise IndexError
        score_col_nums = f"--score-col-nums 3-{3 + betas_shape - 1}"
    except IndexError:
        betas_shape = 1
        betas = betas.reshape(-1, 1)
        score_col_nums = f"--score-col-nums 3"

    # Create the samples file:

    s_table = genotype_matrix.sample_table

    keep_file = osp.join(temp_dir, 'samples.keep')
    keep_table = s_table.get_individual_table()
    keep_table.to_csv(keep_file, index=False, header=False, sep="\t")

    eff_file = osp.join(temp_dir, 'variant_effect_size.txt')
    df = genotype_matrix.get_snp_table(['SNP', 'A1'])

    for i in range(betas_shape):
        df['BETA' + str(i)] = betas[:, i]

    # Remove any variants whose effect size is zero for all traits:
    df = df.loc[df[['BETA' + str(i) for i in range(betas_shape)]].sum(axis=1) != 0]

    # Standardize the genotype, if requested:
    if standardize_genotype:
        standardize_text = ' variance-standardize'
    else:
        standardize_text = ''

    df.to_csv(eff_file, index=False, sep="\t")

    output_file = osp.join(temp_dir, 'samples')

    cmd = [
        f"--bfile {genotype_matrix.bed_file}",
        f"--keep {keep_file}",
        f"--score {eff_file} 1 2 header-read cols=+scoresums{standardize_text}",
        score_col_nums,
        f"--out {output_file}",
    ]

    plink2.execute(cmd)

    if not osp.isfile(output_file + '.sscore'):
        raise FileNotFoundError

    dtypes = {'FID': str, 'IID': str}
    for i in range(betas_shape):
        dtypes.update({'PRS' + str(i): np.float64})

    chr_pgs = pd.read_csv(output_file + '.sscore',
                          delim_whitespace=True,
                          names=['FID', 'IID'] + ['PRS' + str(i) for i in range(betas_shape)],
                          skiprows=1,
                          usecols=[0, 1] + [4 + betas_shape + i for i in range(betas_shape)],
                          dtype=dtypes)
    chr_pgs = keep_table.astype({'FID': str, 'IID': str}).merge(chr_pgs)

    pgs = chr_pgs[['PRS' + str(i) for i in range(betas_shape)]].values

    if betas_shape == 1:
        pgs = pgs.flatten()

    return pgs



from typing import Union
import magenpy as mgp
from magenpy.GWADataLoader import GWADataLoader
from magenpy.SumstatsTable import SumstatsTable
import matplotlib.pylab as plt
import numpy as np


def manhattan(gdl: Union[GWADataLoader, None] = None,
              sumstats: Union[SumstatsTable, None] = None,
              y=None,
              y_label=None,
              chrom_sep_color='#f0f0f0',
              snp_color='#808080',
              snp_marker='o',
              snp_alpha=0.3,
              add_bonf_line=True,
              bonf_line_color='#b06a7a'):

    """
    Generate Manhattan plot where the x-axis is the genomic position (in BP)
    and the y-axis is the -log10(p-value) or some other statistic of the user's choice.

    TODO: Add functionality to highlight certain SNPs or markers on the plot.

    :param gdl: An instance of `GWADataLoader`.
    :param sumstats: An instance of `SumstatsTable`.
    :param y: A vector of values to plot on the y-axis.
    :param y_label: A label for the quantity or statistic that will be plotted.
    :param chrom_sep_color: The color for the chromosome separator block.
    :param snp_color: The color of the dots on the Manhattan plot.
    :param snp_marker: The shape of the marker on the Manhattan plot.
    :param snp_alpha: The opacity level for the markers.
    :param add_bonf_line: If True, add a line indicating the Bonferroni significance threshold.
    :param bonf_line_color: The color of the Bonferroni significance threshold line.

    """

    if y is not None:
        assert y_label is not None

    if gdl is None:
        pos = {c: ss.bp_pos for c, ss in sumstats.split_by_chromosome().items()}
    else:
        pos = {c: ss.bp_pos for c, ss in gdl.sumstats_table.items()}

    starting_pos = 0
    ticks = []
    chrom_spacing = .1*min([p.max() - p.min() for c, p in pos.items()])

    if y is None:
        # If the values for the Y-axis are not provided,
        # we assume that the user wishes to plot a standard Manhattan plot
        # with -log10(p_value) on the Y-axis.

        if add_bonf_line:
            # Add bonferroni significance threshold line:
            plt.axhline(-np.log10(0.05 / 1e6), ls='--', zorder=1, color=bonf_line_color)

        if gdl is None:
            y = {c: ss.log10_p_value for c, ss in sumstats.split_by_chromosome()}
        else:
            y = {c: ss.log10_p_value for c, ss in gdl.sumstats_table.items()}

        y_label = "$-log_{10}$(p-value)"

    if gdl is None:
        unique_chr = sumstats.chromosomes
    else:
        unique_chr = gdl.chromosomes

    for i, c in enumerate(unique_chr):

        min_pos = pos[c].min()
        max_pos = pos[c].max()

        xmin = min_pos + starting_pos
        xmax = max_pos + starting_pos
        if i % 2 == 1:
            plt.axvspan(xmin=xmin, xmax=xmax, zorder=0, color=chrom_sep_color)

        ticks.append((xmin + xmax) / 2)

        plt.scatter(pos[c] + starting_pos, y[c],
                    c=snp_color, alpha=snp_alpha, label=None,
                    marker=snp_marker)

        #if hl_snps is not None:
        #    plt.scatter((pos + starting_pos)[hl_snps[c]], y[c][hl_snps[c]],
        #                c=hl_snp_color, alpha=snp_alpha, label=hl_snp_label,
        #                marker=hl_snp_marker)

        starting_pos += max_pos + chrom_spacing

    plt.xticks(ticks, unique_chr)

    plt.xlabel("Genomic Position")
    plt.ylabel(y_label)

    plt.tight_layout()


def qq_plot(gdl: Union[mgp.GWADataLoader, None] = None,
            sumstats: Union[mgp.SumstatsTable, None] = None,
            statistic='p_value'):
    """
    Generate a quantile-quantile (QQ) plot for the GWAS summary statistics.
    The function supports plotting QQ plots for the -log10(p-values) as well as
    the z-score (if available).

    :param gdl: An instance of `GWADataLoader`.
    :param sumstats: An instance of `SumstatsTable`.
    :param statistic: The statistic to generate the QQ plot for. We currently support `p_value` and `z_score`.
    """

    import scipy.stats as stats

    if statistic == 'p_value':

        if gdl is None:
            p_val = sumstats.log10_p_value
            m = sumstats.m
        else:
            p_val = np.concatenate([ss.log10_p_value for ss in gdl.sumstats_table.values()])
            m = gdl.m

        plt.scatter(-np.log10(np.arange(1, m + 1) / m), np.sort(p_val)[::-1])

        line = np.linspace(0., p_val.max() + 0.1*p_val.max())
        plt.plot(line, line, c='red')
        plt.xlabel("Expected $-log_{10}$(p-value)")
        plt.ylabel("Observed $-log_{10}$(p-value)")

    elif statistic == 'z_score':
        if gdl is None:
            z_scs = sumstats.z_score
        else:
            z_scs = np.concatenate([ss.z_score for ss in gdl.sumstats_table.values()])

        stats.probplot(z_scs, dist="norm", plot=plt)
        plt.show()
    else:
        raise ValueError(f"No QQ plot can be generated for the statistic: {statistic}")


def plot_ld_matrix(ldm: mgp.LDMatrix, row_slice=None, col_slice=None, cmap='OrRd'):
    """
    Plot a heatmap representing the LD matrix or portions of it.

    :param ldm: An instance of `LDMatrix`.
    :param row_slice: A `slice` object indicating which rows to extract from the LD matrix.
    :param col_slice: A `slice` object indicating which columns to extract from the LD matrix.
    :param cmap: The color map for the LD matrix plot.
    """

    if row_slice is None:
        row_slice = slice(ldm.shape[0])

    if col_slice is None:
        col_slice = slice(ldm.shape[0])

    plt.imshow(ldm.to_csr_matrix()[row_slice, col_slice].toarray(), cmap=cmap)
    plt.colorbar()
